import warnings
# warnings.filterwarnings("ignore")

import matplotlib.pyplot as plt
import numpy as np
from astropy.wcs.utils import proj_plane_pixel_scales
from astropy import units as u

# From grizli
def get_wcs_pscale(wcs, unit=u.arcsec):
    """Get correct pscale from a `~astropy.wcs.WCS` object
    Parameters
    ----------
    wcs : `~astropy.wcs.WCS` or `~astropy.io.fits.Header`
    unit: `~astropy.units.Unit`
    Returns
    -------
    pscale : float
        Pixel scale
    """
    pscale = proj_plane_pixel_scales(wcs)[0] * 3600 * unit
    return pscale.to(unit).value

def transform_wcs(in_wcs, translation=[0., 0.], rotation=0., scale=1.):
    """Update WCS with shift, rotation, & scale
    Parameters
    ----------
    in_wcs: `~astropy.wcs.WCS`
        Input WCS
    translation: [float, float]
        xshift & yshift in pixels
    rotation: float
        CCW rotation (towards East), radians
    scale: float
        Pixel scale factor
    Returns
    -------
    out_wcs: `~astropy.wcs.WCS`
        Modified WCS
    """
    out_wcs = in_wcs.deepcopy()

    # Compute shift for crval
    crval = in_wcs.all_pix2world([in_wcs.wcs.crpix-np.array(translation)],
                                    1).flatten()

    # Compute shift at image center
    if hasattr(in_wcs, '_naxis1'):
        refpix = np.array([in_wcs._naxis1/2., in_wcs._naxis2/2.])
    else:
        refpix = np.array(in_wcs._naxis)/2.

    c0 = in_wcs.all_pix2world([refpix], 1).flatten()
    c1 = in_wcs.all_pix2world([refpix-np.array(translation)], 1).flatten()

    out_wcs.wcs.crval += c1-c0

    theta = -rotation
    _mat = np.array([[np.cos(theta), -np.sin(theta)],
                    [np.sin(theta), np.cos(theta)]])

    out_wcs.wcs.pc = np.dot(out_wcs.wcs.pc, _mat)/scale

    out_wcs.pscale = get_wcs_pscale(out_wcs)
    if hasattr(out_wcs, 'pixel_shape'):
        _naxis1 = int(np.round(out_wcs.pixel_shape[0]*scale))
        _naxis2 = int(np.round(out_wcs.pixel_shape[1]*scale))
        out_wcs._naxis = [_naxis1, _naxis2]
    elif hasattr(out_wcs, '_naxis1'):
        out_wcs._naxis1 = int(np.round(out_wcs._naxis1*scale))
        out_wcs._naxis2 = int(np.round(out_wcs._naxis2*scale))

    return out_wcs



import warnings
# warnings.filterwarnings("ignore")
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.nddata.utils import Cutout2D
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales
from reproject import reproject_adaptive, reproject_exact, reproject_interp

from .wcs import transform_wcs


def zscale(img):
    """
    Normalization of zscale.
    Input is a 2D numpy array.

    Return a Normalization class to be used with Matplotlib.
    """
    from astropy.visualization import (ImageNormalize, LinearStretch,
                                       ZScaleInterval)
    norm = ImageNormalize(img, interval=ZScaleInterval(),
                          stretch=LinearStretch())
    return norm

def gamma_correction(colorbar, gamma=1.0):
    '''
    Gamma correction for colorbar.
    Input is a colormap instance.

    Return a new colormap instance.
    '''
    from matplotlib import cm
    from matplotlib.colors import ListedColormap
    default = cm.get_cmap(colorbar, 256)
    de_color = default(np.linspace(0, 1, 256))
    newcolors = de_color.copy()
    for i in range(256):
        newcolors[i, :] = de_color[int(255 * np.power(i/255, gamma)), :]
    return ListedColormap(newcolors)

# build RGB image.
def combine_RGB(b, g, r, fwhm=2, Q=10, stretch=0.05, filename=None):
    from astropy.convolution import Gaussian2DKernel, convolve
    from astropy.stats import gaussian_fwhm_to_sigma
    from astropy.visualization import make_lupton_rgb

    sigma = fwhm * gaussian_fwhm_to_sigma
    kernel = Gaussian2DKernel(sigma, x_size=11, y_size=11)
    g = convolve(g, kernel, normalize_kernel=True)
    r = convolve(r, kernel, normalize_kernel=True)
    b = convolve(b, kernel, normalize_kernel=True)

    if(filename):
        rgb = make_lupton_rgb(r, g, b, Q=Q, stretch=stretch, filename=filename)
    else:
        rgb = make_lupton_rgb(r, g, b, Q=Q, stretch=stretch)

    return rgb

def plot_beam(ax, header, xy=(2,2)):
    import matplotlib.patches as mpatches
    BMAJ = 3600. * header["BMAJ"] # [arcsec]
    BMIN = 3600. * header["BMIN"] # [arcsec]
    BPA =  header["BPA"] # degrees East of North
    print('BMAJ: {:.3f}", BMIN: {:.3f}", BPA: {:.2f} deg'.format(BMAJ, BMIN, BPA))
    # However, to plot it we need to negate the BPA since the rotation is the opposite direction
    # due to flipping RA.
    ax.add_artist(mpatches.Ellipse(xy=xy, width=BMIN, height=BMAJ, angle=-BPA, facecolor="none", color='white',linewidth=3))



class AstroImage:
    def __init__(self, data, header):
        self.data = data
        self.header = header
        self.wcs = WCS(self.header)

    @classmethod
    def read(cls, url, ext=0):
        with fits.open(url) as hdulist:
            data = hdulist[ext].data
            header = hdulist[ext].header
        return cls(data, header)

    @property
    def pixel_scale(self, unit=u.arcsec):
        pscale = proj_plane_pixel_scales(self.wcs)[0] * 3600 * unit
        return pscale.to(unit).value

    @property
    def shape(self):
        x = getattr(self.wcs, 'NAXIS1', np.shape(self.data)[1])
        y = getattr(self.wcs, 'NAXIS2', np.shape(self.data)[0])
        return x, y

    @property
    def skycenter(self):
        ra, dec = np.array(self.wcs.wcs_pix2world(*self.pixcenter, 0))
        return ra, dec

    @property
    def pixcenter(self):
        x, y = self.shape
        x, y = (x-1)/2, (y-1)/2
        return x, y

    @property
    def hdu(self):
        return fits.ImageHDU(data=self.data, header=self.header)

    @property
    def footprint(self):
        return AstroImage(data=(~np.isnan(self.data)).astype(float), header=self.header)

    def __repr__(self, plot=True):
        __info__ = """
        Image information:
        ------------------
        Image shape: {} pixels <-> {:.3f} x {:.3f} arcsec
        Image center: {:.7f}, {:.7f} (RA, Dec)
        Pixel scale: {:.3f} arcsec/pixel
        """.format(self.shape, self.shape[0] * self.pixel_scale, self.shape[1] * self.pixel_scale,
        self.skycenter[0], self.skycenter[1], self.pixel_scale)

        if(plot):
            self.preview()
        return __info__

    def save(self, url):
        hdu = fits.PrimaryHDU(self.data, header=self.header)
        hdu.writeto(url, overwrite=True)

    def preview(self, color_map='gray_r', gamma=1.0, colorbar=True, **kwargs):
        fig, ax = plt.subplots(figsize=(6, 6), dpi=100, subplot_kw={'projection': self.wcs})
        norm = zscale(self.data)
        if(gamma == 1):
            img = ax.imshow(self.data, cmap=color_map,
                            norm=norm, origin='lower')
        else:
            img = ax.imshow(self.data, cmap=gamma_correction(
                color_map, gamma=gamma), norm=norm, origin='lower')
        ax.set_xlabel('RA')
        ax.set_ylabel('Dec')
        if(colorbar):
            fig.colorbar(img, ax=ax, shrink=0.8)
        return fig, ax

    def cutout(self, coord, size, coord_unit=u.deg, size_unit=u.arcsec, mode='partial'):
        if(coord_unit == u.deg):
            coord = SkyCoord(coord[0], coord[1], unit='deg')
        if(size_unit == u.arcsec):
            size = size * size_unit
        elif(size_unit == 'pixel'):
            size = size
        hdu_crop = Cutout2D(self.data, position=coord, size=size,
                            wcs=self.wcs, mode=mode)
        wcs_crop = hdu_crop.wcs
        self.header.update(wcs_crop.to_header())
        return AstroImage(data=hdu_crop.data, header=self.header)

    def rotate(self, angle, algorithm='interpolation'):
        input_wcs = deepcopy(self.wcs)
        input_wcs.wcs.crpix = self.pixcenter
        input_wcs.wcs.crval = self.skycenter
        output_wcs = transform_wcs(input_wcs, rotation=np.deg2rad(angle))
        if(algorithm == 'interpolation'):
            data = reproject_interp(self.hdu, output_wcs, shape_out=np.array(self.shape), return_footprint=False)
        elif(algorithm == 'exact'):
            data = reproject_exact(self.hdu, output_wcs, shape_out=np.array(self.shape), return_footprint=False)
        elif(algorithm == 'adaptive'):
            data = reproject_adaptive(self.hdu, output_wcs, shape_out=np.array(self.shape), return_footprint=False, conserve_flux=True)
        else:
            raise ValueError('Algorithm not supported')
        return AstroImage(data=data, header=output_wcs.to_header())

    def mask_blank(self, threshold=10000):
        '''
        Masking the probably border or blank region
        '''
        values, counts = np.unique(self.data, return_counts=True)
        critical_counts = np.mean(counts) + 5*np.std(counts) + threshold
        print("critical counts = ", critical_counts)
        mode_values = values[counts > critical_counts]
        mode_counts = counts[counts > critical_counts]
        for i, m in enumerate(mode_values):
            print(mode_counts[i], " pixels' value is ", m)
            outregion = (self.data == m)
            self.data[outregion] = np.nan
        return self



# -*- encoding: utf-8 -*-
'''
@File    :   scampy.py
@Time    :   2022/09/10
@Author  :   Mingyu Li
@Contact :   lmytime@hotmail.com
'''


from astropy.io import fits
from astropy.wcs import WCS
from astrolib.astro_image import AstroImage
import getpass
import subprocess
import os, time

def dict_to_command_string(config):
    return "".join([ f" -{key} {value} " for key, value in config.items() ])

class Sextrapy:
    def __init__(self, file, folder="."):
        '''
            Initialize the class, specify the folder to store the files
        '''
        self.dir = folder
        self.file = file
        os.makedirs(self.dir, exist_ok=True)

    @classmethod
    def config_sextractor_for_scamp(self):
        sextractor_config = {
            "CATALOG_TYPE": "FITS_LDAC",
            "DETECT_THRESH": 3.0,
            "ANALYSIS_THRESH": 3.0,
            "FILTER": "N",
            "PIXEL_SCALE": 0.0, # size of pixel in arcsec (0=use FITS WCS info)
            "MAG_ZEROPOINT": 0.0,
        }
        return sextractor_config

    @classmethod
    def config_sextractor(self):
        sextractor_config = {
            "CATALOG_NAME": "test.cat",
            "CATALOG_TYPE": "ASCII_HEAD",
            "PARAMETERS_NAME": "default.param",
            "DETECT_TYPE": "CCD",
            "DETECT_MINAREA": 5,
            "DETECT_THRESH": 3,
            "ANALYSIS_THRESH": 3,
            "FILTER": "Y",
            "FILTER_NAME": "default.conv",
            "DEBLEND_NTHRESH": 32,
            "DEBLEND_MINCONT": 0.005,
            "CLEAN": "Y",
            "CLEAN_PARAM": 1.0,
            "MASK_TYPE": "CORRECT",
            "PHOT_APERTURES": 5,
            "PHOT_AUTOPARAMS": "2.5,3.5",
            "PHOT_PETROPARAMS": "2.0,3.5",
            "SATUR_LEVEL": 50000.0,
            "SATUR_KEY": "SATURATE",
            "MAG_ZEROPOINT": 0.0,
            "MAG_GAMMA": 4.0,
            "GAIN": 0.0,
            "GAIN_KEY": "GAIN",
            "PIXEL_SCALE": 0.0,
            "SEEING_FWHM": 1.2,
            "STARNNW_NAME": "default.nnw",
            "BACK_SIZE": 64,
            "BACK_FILTERSIZE": 3,
            "BACKPHOTO_TYPE": "GLOBAL",
            "CHECKIMAGE_TYPE": "NONE",
            "CHECKIMAGE_NAME": "check.fits",
            "MEMORY_OBJSTACK": 3000,
            "MEMORY_PIXSTACK": 300000,
            "MEMORY_BUFSIZE": 1024,
            "VERBOSE_TYPE": "NORMAL",
            "HEADER_SUFFIX": ".head",
            "WRITE_XML": "N",
        }
        return sextractor_config

    def paramters_sextractor_for_scamp(self):
        '''
            Write a paramters_sextractor_for_scamp.param file for sextractor
        '''
        params = ["NUMBER", "FLUX_AUTO", "FLUXERR_AUTO", "ALPHAWIN_J2000", "DELTAWIN_J2000",
                    "ERRAWIN_WORLD", "ERRBWIN_WORLD", "ERRTHETA_WORLD",
                    "XWIN_IMAGE", "YWIN_IMAGE", "ERRAWIN_IMAGE", "ERRBWIN_IMAGE",
                    "ERRTHETAWIN_IMAGE", "MAG_AUTO", "MAGERR_AUTO", "FLAGS", "FLAGS_WEIGHT"]
        params = "\n".join(params)
        params = "#This file is generated by astrolib.astromatic module\n#      (c) Mingyu\n# Parameters prepared for SCAMP\n" + params
        with open(os.path.join(self.dir, "paramters_sextractor_for_scamp.param"), "w") as f:
            f.write(params)

    def run_sextractor(self, sextractor="sex", extension=None, config=None):
        # sepcify sextractor configure
        if config is None:
            sextractor_config = self.config_sextractor()
        else:
            sextractor_config = config

        # specify extension of fits file to be processed
        if extension is None:
            extension_string = ""
        else:
            extension_string = f"[{int(extension)}]"

        # run sextractor
        config_string = dict_to_command_string(sextractor_config)
        command_string = f"{sextractor} {self.file}{extension_string} {config_string}"
        print(command_string)
        subprocess.call(command_string, shell=True)


class Scampy:

    def __init__(self, file, folder="."):
        '''
            Initialize the class, specify the folder to store the files
        '''
        self.dir = folder
        self.file = file
        os.makedirs(self.dir, exist_ok=True)

    def run_sextractor_for_scamp(self, sextractor="sex", extension=None, config=None):
        sex = Sextrapy(self.file, self.dir)
        # sepcify sextractor configure
        if config is None:
            config = sex.config_sextractor_for_scamp()
        config['CATALOG_NAME'] = os.path.join(self.dir, self.file.replace('.fits', '.LDAC'))
        config['PARAMETERS_NAME'] = os.path.join(self.dir, 'paramters_sextractor_for_scamp.param')
        sex.paramters_sextractor_for_scamp()
        sex.run_sextractor(sextractor=sextractor, extension=None, config=config)

    @classmethod
    def config_scamp(self):
        scamp_config = {
            'REF_SERVER': 'vizier.nao.ac.jp', # Internet addresses of catalog servers
                                                # Possible mirrors include:
                                                # vizier.nao.ac.jp,
                                                # vizier.hia.nrc.ca,
                                                # vizier.ast.cam.ac.uk,
                                                # vizier.iucaa.in,
                                                # vizier.china-vo.org,
                                                # vizier.cfa.harvard.edu and
                                                # viziersaao.chpc.ac.za
            'ASTREF_CATALOG': 'GAIA-DR2',   # NONE,FILE,USNO-A2,USNO-B1,GSC-2.3,
                                                # TYCHO-2,UCAC-4,URAT-1,NOMAD-1,PPMX,
                                                # CMC-15,2MASS,DENIS-3,SDSS-R9,SDSS-R12,
                                                # IGSL,GAIA-DR1,GAIA-DR2,GAIA-EDR3,
                                                # PANSTARRS-1, or ALLWISE
            'ASTREF_BAND': 'DEFAULT',
            'ASTREFCAT_NAME': 'astrefcat.cat',
            'ASTREFCENT_KEYS': 'ALPHAWIN_J2000,DELTAWIN_J2000',
            'ASTREFERR_KEYS': 'ERRAWIN_WORLD,ERRBWIN_WORLD,ERRTHETAWIN_WORLD',
            'ASTREFMAG_KEY': 'MAG_AUTO',
            'ASTREFMAGERR_KEY': 'MAGERR_AUTO',
            'ASTREFMAG_LIMITS': '16.0,19.0', # Select magnitude range in ASTREF_BAND
            'CROSSID_RADIUS': '1.0',
            'DISTORT_DEGREES': '2',
            'SOLVE_PHOTOM': 'N',
            'SN_THRESHOLDS': '20.0,100.0',   # S/N thresholds (in sigmas) for all and
                                                # high-SN sample
            'AHEADER_SUFFIX': '.ahead',
            'HEADER_SUFFIX': '.head',
            'WRITE_XML': 'N',
        }
        return scamp_config

    def run_scamp(self, scamp="scamp", config=None):
        # sepcify sextractor configure
        if config is None:
            scamp_config = self.config_scamp()
        else:
            scamp_config = config

        # run scamp
        config_string = dict_to_command_string(scamp_config)
        command_string = f"{scamp} {os.path.join(self.dir, self.file.replace('.fits', '.LDAC'))} {config_string}"
        print(command_string)
        subprocess.call(command_string, shell=True)

    @classmethod
    def read_scamp_header(self, header=None):
        '''
            Read the header file generated by scamp
        '''
        if header is None:
            header = os.path.join(self.dir, self.file.replace('.fits', '.head'))
        with open(header, "r") as f:
            headers = f.readlines()
        new_header = fits.Header.fromstring("".join(headers[3:17]))
        return new_header

    @classmethod
    def update_header(self, header, new_header):
        '''
            Update the header of the original header
        '''

        for card in new_header.cards:
            header[card[0]] = card[1:]
        header['history'] = f"WCS calibrated with SCAMP by {getpass.getuser()} on {time.strftime('%Y/%m/%d %H:%m', time.localtime())}"

        return header

    @classmethod
    def update_header_for_fits(self, extension, new_header, file=None):
        '''
            Update the header of the original fits file
        '''
        if file is None:
            file = self.file
        with fits.open(file) as hdul:
            header = hdul[extension].header
            hdul[extension].header = self.update_header(header, new_header)
            hdul.writeto(file.replace('.fits', '_newwcs.fits'), overwrite=True)


import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u
from astropy.stats import sigma_clip
from astropy.convolution import convolve,Gaussian2DKernel
from spectral_cube import SpectralCube
from scipy import ndimage
import regions
from astropy import units as u
from astropy.coordinates import SkyCoord

class AstroCube(SpectralCube):

    def __init__(self, cube, **krg):
        '''
        初始化cube实例
        '''
        # super().__init__(**krg) #因为继承自SpectralCube,这句是初始化父类
        self.cube = cube

    @classmethod#类方法
    def read(cls,filename,**krg):
        '''
        class function, read the datacube from .fits
        '''
        cube = super().read(filename=filename, **krg)
        # cube = SpectralCube(data=cube.hdu.data, wcs=cube.wcs, **krg)
        return AstroCube(cube)

    def aper_spec(self, coord, radius):
        '''
        从cube中提取一个圆形区域的光谱
        coord:中心坐标
        radius:半径
        '''
        regpix = regions.CircleSkyRegion(center=SkyCoord(coord[0], coord[1], unit='deg'), radius=radius*u.arcsec)
        subcube = self.cube.subcube_from_regions([regpix])
        spectrum = subcube.sum(axis=(1,2))
        return spectrum


# import warnings
# warnings.filterwarnings("ignore")

import matplotlib.pyplot as plt
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.nddata.utils import Cutout2D
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales
from matplotlib import cm
from matplotlib.colors import ListedColormap
from photutils.aperture import SkyCircularAperture, aperture_photometry, SkyCircularAnnulus
from photutils.aperture import ApertureStats
from astropy import stats
from astropy.stats import SigmaClip




def zscale(img):
    """
    Normalization of zscale.
    Input is a 2D numpy array.

    Return a Normalization class to be used with Matplotlib.
    """
    from astropy.visualization import ImageNormalize, LinearStretch, ZScaleInterval
    norm = ImageNormalize(img, interval=ZScaleInterval(),
                          stretch=LinearStretch())
    return norm

def gamma_correction(colorbar, gamma=1.0):
    '''
    Gamma correction for colorbar.
    Input is a colormap instance.

    Return a new colormap instance.
    '''
    default = cm.get_cmap(colorbar, 256)
    de_color = default(np.linspace(0, 1, 256))
    newcolors = de_color.copy()
    for i in range(256):
        newcolors[i, :] = de_color[int(255 * np.power(i/255, gamma)), :]
    return ListedColormap(newcolors)

# build RGB image.
def combine_RGB(b, g, r, fwhm=2, Q=10, stretch=0.05, filename=None):
    from astropy.visualization import make_lupton_rgb
    from astropy.convolution import Gaussian2DKernel, convolve
    from astropy.stats import gaussian_fwhm_to_sigma

    sigma = fwhm * gaussian_fwhm_to_sigma
    kernel = Gaussian2DKernel(sigma, x_size=10, y_size=10)
    g = convolve(g, kernel, normalize_kernel=True)
    r = convolve(r, kernel, normalize_kernel=True)
    b = convolve(b, kernel, normalize_kernel=True)

    if(filename):
        rgb = make_lupton_rgb(r, g, b, Q=Q, stretch=stretch, filename=filename)
    else:
        rgb = make_lupton_rgb(r, g, b, Q=Q, stretch=stretch)

    return rgb

def plot_beam(ax, header, xy=(2,2)):
    import matplotlib.patches as mpatches
    BMAJ = 3600. * header["BMAJ"] # [arcsec]
    BMIN = 3600. * header["BMIN"] # [arcsec]
    BPA =  header["BPA"] # degrees East of North
    print('BMAJ: {:.3f}", BMIN: {:.3f}", BPA: {:.2f} deg'.format(BMAJ, BMIN, BPA))
    # However, to plot it we need to negate the BPA since the rotation is the opposite direction
    # due to flipping RA.
    ax.add_artist(mpatches.Ellipse(xy=xy, width=BMIN, height=BMAJ, angle=-BPA, facecolor="none", color='white',linewidth=3))


class AstroImage:
    def __init__(self):
        pass

    @classmethod
    def read(cls, url, ext):
        with fits.open(self.url) as hdulist:
            self.hdu = hdulist[self.ext].data
            self.hdr = hdulist[self.ext].header
            self.wcs = WCS(self.hdr)

class AstroImage:
    def __init__(self, url, ext=0, mode='file', verbose=False, **kwargs):
        '''
        初始化AstroImage
        '''
        self.url = url
        self.ext = ext
        if(mode == 'file'):
            self.open_fits_image()
            # 判断是否有多余的维度
            if((self.wcs.wcs.naxis != 2) & verbose):
                print(
                    f"This is not a 2D image.\nThis image has {self.wcs.wcs.naxis} axises: {self.wcs.wcs.ctype}.\nTry to auto drop axises.")
            while((self.wcs.wcs.naxis != 2) and (self.wcs.array_shape[0] == 1)):
                self.hdu = self.hdu[0]
                self.wcs = self.wcs.dropaxis(dropax=self.wcs.wcs.naxis-1)
                self.hdr.update(self.wcs.to_header())
            if((self.wcs.wcs.naxis == 2) & verbose):
                print(
                    f"Drop axises success! Now image has {self.wcs.wcs.naxis} axises: {self.wcs.wcs.ctype}.")
            else:
                Exception(
                    f"Drop axixes ERROR!\nThis image still has {self.wcs.wcs.naxis} axises: {self.wcs.wcs.ctype}.")

            # Pixel scale
            self.pixel_scale = proj_plane_pixel_scales(self.wcs) * 3600
        elif(mode == 'data'):
            self.hdu = kwargs['hdu']
            self.hdr = kwargs['hdr']
            self.wcs = kwargs['wcs']
            self.pixel_scale = kwargs['pixel_scale']

    def open_fits_image(self):
        '''
        初始化时候用到的读取文件的函数
        '''
        with fits.open(self.url) as hdulist:
            self.hdu = hdulist[self.ext].data
            self.hdr = hdulist[self.ext].header
            self.wcs = WCS(self.hdr)

    def cutout(self, coord, box):
        '''
        截图的函数，coord输入的ra，dec的数组；box输入的矩形长和宽，单位是arcsec
        '''
        coord = np.array([coord], dtype=np.float64)
        x, y = self.wcs.wcs_world2pix(coord, 0)[0]
        box = np.array(box) / np.array(self.pixel_scale)
        hdu_crop = Cutout2D(self.hdu, (x, y), box,
                            wcs=self.wcs, mode='partial')
        wcs_crop = hdu_crop.wcs
        self.hdr.update(wcs_crop.to_header())
        return AstroImage(url="", mode="data", hdu=hdu_crop.data, hdr=self.hdr, wcs=wcs_crop, pixel_scale=self.pixel_scale)

    def cutout_pixel(self, coord, box):
        '''
        截图的函数，coord输入的ra，dec的数组；box输入的矩形长和宽，单位是pixel
        '''
        coord = np.array([coord], dtype=np.float64)
        x, y = self.wcs.wcs_world2pix(coord, 0)[0]
        box = np.array(box)
        hdu_crop = Cutout2D(self.hdu, (x, y), box,
                            wcs=self.wcs, mode='partial')
        wcs_crop = hdu_crop.wcs
        self.hdr.update(wcs_crop.to_header())
        return AstroImage(url="", mode="data", hdu=hdu_crop.data, hdr=self.hdr, wcs=wcs_crop, pixel_scale=self.pixel_scale)

    def preview(self, color_map='gray_r', gamma=1.0, colorbar=True, **kwargs):
        '''
        预览图像，只做预览用，科学图像应手动画。
        '''
        fig, ax = plt.subplots(figsize=(6, 6), dpi=100,
                               subplot_kw={'projection': self.wcs})
        norm = zscale(self.hdu)
        if(gamma == 1):
            img = ax.imshow(self.hdu, cmap=color_map,
                            norm=norm, origin='lower')
        else:
            img = ax.imshow(self.hdu, cmap=gamma_correction(
                color_map, gamma=gamma), norm=norm, origin='lower')
        ax.set_xlabel('RA')
        ax.set_ylabel('Dec')
        if(colorbar):
            fig.colorbar(img, ax=ax, shrink=0.8)
        return fig, ax

    def save(self, url):
        '''
        保存为fits文件
        '''
        hdu = fits.PrimaryHDU(self.hdu, header=self.hdr)
        hdu.writeto(url, overwrite=True)

    def __repr__(self):
        '''
        在输出的时候显示一些信息
        '''
        fig, ax = self.preview()
        plt.show()
        return f"2D {self.wcs.array_shape} AstroImage"

    def mask_blank(self, threshold=10000):
        '''
        Masking the probably border or blank region
        '''
        values, counts = np.unique(self.hdu, return_counts=True)
        critical_counts = np.mean(counts) + 5*np.std(counts) + threshold
        print("critical counts = ", critical_counts)
        mode_values = values[counts > critical_counts]
        mode_counts = counts[counts > critical_counts]
        for i, m in enumerate(mode_values):
            print(mode_counts[i], " pixels' value is ", m)
            outregion = (self.hdu == m)
            self.hdu[outregion] = np.nan
        return self

    def photometry(self, coord, r=1, an_r=[1,2], mode='pixel', zeropoint=0, plot=False):
        '''
        圆形口径测光
        '''
        if(mode == 'pixel'):
            positions = self.wcs.pixel_to_world(*coord)
        elif(mode == 'sky'):
            positions = SkyCoord(
                ra=coord[0]*u.degree, dec=coord[1]*u.degree, frame='fk5')
        aperture = SkyCircularAperture(positions, r=r*u.arcsec)
        annulus_aperture = SkyCircularAnnulus(positions, r_in=an_r[0]*u.arcsec, r_out=an_r[1]*u.arcsec)
        sigclip = SigmaClip(sigma=3.0, maxiters=10)
        aper_stats = ApertureStats(self.hdu, aperture, sigma_clip=None, wcs=self.wcs)
        bkg_stats = ApertureStats(self.hdu, annulus_aperture, sigma_clip=sigclip, wcs=self.wcs)
        total_bkg = bkg_stats.median * aper_stats.sum_aper_area.value
        counts_err = bkg_stats.std * np.sqrt(aper_stats.sum_aper_area.value)
        apersum_bkgsub = aper_stats.sum - total_bkg
        counts = apersum_bkgsub
        if(counts > 2*counts_err):
            mag = -2.5 * np.log10(counts) + zeropoint
            mag_err = 2.5/np.log(10) * counts_err / counts
        else:
            mag = -2.5 * np.log10(2*counts_err) + zeropoint
            mag_err = 99
        if(not plot):
            return mag, mag_err, aperture
        else:
            fig, ax = self.preview()
            ap_patches = aperture.to_pixel(self.wcs).plot(color='red', lw=2)
            ax.text(0.05, 0.9, f"{mag=:.2f}\n{r=:.2f}''",
                    transform=plt.gca().transAxes, color='red')
            return mag, mag_err, aperture, fig, ax

    def sigma_clipped_stats(self, sigma=2, maxiters=5, **kwargs):
        '''
        圆形口径测光
        '''
        mean, median, stddev = stats.sigma_clipped_stats(self.hdu, sigma=sigma, maxiters=maxiters, **kwargs)
        return mean, median, stddev

    def mag_err(self, mag, zeropoint, area):
        '''
        计算magnitude的误差
        '''
        mean, median, stddev = self.sigma_clipped_stats()
        flux_err = stddev * np.sqrt(area)
        mag_err = 2.5/np.log(10) * flux_err / 10**((mag-zeropoint)/(-2.5))
        return mag_err


import astropy.units as u
from astropy.coordinates import SkyCoord

def catalog_match(ra1, dec1, ra2, dec2):
    c1 = SkyCoord(ra=ra1*u.degree, dec=dec1*u.degree)
    c2 = SkyCoord(ra=ra2*u.degree, dec=dec2*u.degree)
    idx, d2d, _ = c1.match_to_catalog_sky(c2)

    return idx, d2d

def to_ds9_region(ra, dec, text=None, filename="ds9.reg"):
    with open(filename,'w+',newline='') as txtfile:
        txtfile.write('# Region file format: DS9 version 4.1\n')
        txtfile.write('global color=green dashlist=8 3 width=1 font="helvetica 10 normal roman" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1\n')
        txtfile.write('fk5\n')
        if text is None:
            for _ra, _dec in zip(ra, dec):
                txtfile.write(f"point({_ra},{_dec}) # point=diamond width=2 color=cyan \n")
        else:
            for _ra, _dec, _text in zip(ra, dec, text):
                txtfile.write(f"point({_ra},{_dec}) # point=diamond width=2 color=cyan")
                txtfile.write(" text={" + str(_text) + "}\n")